package org.greenstone.gsdl3.util;

import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.w3c.dom.Document;
import org.w3c.dom.Text;

import java.util.Map;
import java.util.Set;
import java.util.HashMap;
import java.util.Vector;
import java.util.Iterator;
//import java.util.Locale;

/** various functions for extracting info out of GS XML */
public class GSXML {

    // greenstone xml elements
    public static final String MESSAGE_ELEM = "message";
    public static final String REQUEST_ELEM = "request";
    public static final String RESPONSE_ELEM = "response";
    public static final String COLLECTION_ELEM = "collection";
    public static final String SERVICE_ELEM = "service";
    public static final String CLUSTER_ELEM = "serviceCluster";
    public static final String SITE_ELEM = "site";
    public static final String PARAM_ELEM = "param";
    public static final String PARAM_OPTION_ELEM = "option";
    public static final String CONTENT_ELEM = "content";
    public static final String RESOURCE_ELEM = "resource";
    public static final String DOCUMENT_ELEM = "document";
    public static final String METADATA_ELEM = "metadata";
    public static final String SERVICE_CLASS_ELEM = "serviceRack";
    public static final String CLASSIFIER_ELEM = "classifier";
    public static final String APPLET_ELEM = "applet";
    public static final String APPLET_DATA_ELEM = "appletData";
    public static final String CONFIGURE_ELEM = "configure";
    public static final String STATUS_ELEM = "status";
    public static final String ERROR_ELEM = "error";
    public static final String DEFAULT_ELEM = "default";
    public static final String STYLESHEET_ELEM = "format";//"stylesheet"; // any additional stylesheet stuff is carried in the message inside this elem
    public static final String FORMAT_ELEM = "format"; // config files use format - shoudl we use this instead of stylesheet??
    public static final String TERM_ELEM = "term";
    public static final String SYSTEM_ELEM = "system";

    //config file elems
    public static final String INDEX_ELEM = "index";
    public static final String SEARCH_ELEM = "search";
    public static final String BROWSE_ELEM = "browse";
    public static final String DISPLAY_ELEM = "display";
    public static final String FIELD_ELEM = "field";
    public static final String SHORTNAME_ATT = "shortname";

    // elems for the pages to be processed by xslt
    public final static String PAGE_ELEM = "page";
    public final static String CONFIGURATION_ELEM = "config";
    public final static String PAGE_REQUEST_ELEM = "pageRequest";
    public final static String PAGE_RESPONSE_ELEM = "pageResponse";
    public final static String PAGE_EXTRA_ELEM = "pageExtra";

    //public final static String DESCRIPTION_ELEM = "description";

    public static final String SITE_NAME_ELEM = "localSiteName";
    public static final String SITE_HTTP_ADDRESS_ELEM = "httpAddress";
    public static final String  ACTION_ELEM = "action";
    public static final String  SUBACTION_ELEM = "subaction";
    
    // add on to another elem type to get a list of that type
    public static final String LIST_MODIFIER = "List";

    // greenstone xml attributes
    public static final String NAME_ATT = "name";
    public static final String TO_ATT = "to";
    public static final String FROM_ATT = "from";
    public static final String LANG_ATT = "lang";
    public static final String TYPE_ATT = "type";
    public static final String VALUE_ATT = "value";
    public static final String DEFAULT_ATT = "default";
    public static final String INFO_ATT = "info";
    public static final String ACTION_ATT = "action";
    public static final String SUBACTION_ATT = "subaction";
    public static final String OUTPUT_ATT = "output";
    public static final String ADDRESS_ATT = "address";
    public static final String STATUS_ERROR_CODE_ATT = "code";
    public static final String STATUS_PROCESS_ID_ATT = "pid";
    public static final String PARAM_SHORTNAME_ATT = "shortname";
    public static final String PARAM_IGNORE_POS_ATT = "ignore";
    public static final String CLASSIFIER_CONTENT_ATT = "content";

    // document stuff
    public static final String DOC_TYPE_ATT = "docType";
    public static final String DOC_NODE_ELEM = "documentNode";
    public static final String NODE_CONTENT_ELEM = "nodeContent";
    public static final String NODE_STRUCTURE_ELEM = "nodeStructure";
    public static final String NODE_ID_ATT = "nodeID";
    public static final String NODE_NAME_ATT = "nodeName";
    public static final String NODE_TYPE_ATT = "nodeType";

    public static final String NODE_TYPE_ROOT = "root";
    public static final String NODE_TYPE_INTERIOR = "interior";
    public static final String NODE_TYPE_LEAF = "leaf";

    // classifier stuff
    public static final String CLASS_NODE_ELEM = "classifierNode";
    public static final String CLASS_NODE_ORIENTATION_ATT = "orientation";
    
    // parameter types
    public static final String PARAM_TYPE_INTEGER = "integer";
    public static final String PARAM_TYPE_BOOLEAN = "boolean";
    public static final String PARAM_TYPE_ENUM_START = "enum";
    public static final String PARAM_TYPE_ENUM_SINGLE = "enum_single";
    public static final String PARAM_TYPE_ENUM_MULTI = "enum_multi";
    public static final String PARAM_TYPE_STRING = "string";
    public static final String PARAM_TYPE_TEXT = "text";
    public static final String PARAM_TYPE_MULTI = "multi";
    public static final String PARAM_TYPE_FILE = "file";
    public static final String PARAM_TYPE_INVISIBLE = "invisible";
    // stuff for text strings 
    public static final String DISPLAY_TEXT_ELEM = "displayItem";
    // the following are used for the name attributes
    public static final String DISPLAY_TEXT_NAME = "name";
    public static final String DISPLAY_TEXT_SUBMIT = "submit";
    public static final String DISPLAY_TEXT_DESCRIPTION = "description";

    // request types - do we need all these? or just
    // describe, status, and one to say do teh request
    public static final String REQUEST_TYPE_DESCRIBE = "describe";
    public static final String REQUEST_TYPE_STATUS = "status";
    public static final String REQUEST_TYPE_PROCESS = "process";
    public static final String REQUEST_TYPE_CONFIGURE = "configure"; 
    public static final String REQUEST_TYPE_SYSTEM = "system"; 
    public static final String REQUEST_TYPE_PAGE = "page"; // used to be cgi
    public static final String REQUEST_TYPE_FORMAT = "format"; 

    public static final String REQUEST_TYPE_ACTION = "action"; // ??
    public static final String REQUEST_TYPE_QUERY = "query"; // ??
    public static final String REQUEST_TYPE_BUILD = "build"; // ??
   
    // service types
    public static final String SERVICE_TYPE_QUERY = "query";
    public static final String SERVICE_TYPE_RETRIEVE = "retrieve";
    public static final String SERVICE_TYPE_BROWSE = "browse";
    public static final String SERVICE_TYPE_APPLET = "applet";
    public static final String SERVICE_TYPE_PROCESS = "process";
    public static final String SERVICE_TYPE_ENRICH = "enrich";

    // configure types - get rid of these first two
    public static final String CONFIG_ACTION_ACTIVATE = "activate";
    public static final String CONFIG_ACTION_DEACTIVATE = "deactivate";

    public static final String SYSTEM_TYPE_CONFIGURE = "configure";
    public static final String SYSTEM_TYPE_ACTIVATE = "activate";
    public static final String SYSTEM_TYPE_DEACTIVATE = "deactivate";

    public static final String SYSTEM_SUBSET_ATT = "subset";
    public static final String SYSTEM_MODULE_TYPE_ATT = "moduleType";
    public static final String SYSTEM_MODULE_NAME_ATT = "moduleName";
    
    // communicator types
    public static final String COMM_TYPE_SOAP_JAVA = "soap";
    
    // some system wide param names
    public static final String SUBSET_PARAM = "subset";

    /** takes a list of elements, and returns an array of strings 
     * of the values of attribute att_name */
    public static String [] getAttributeValuesFromList(Element list,
						       String att_name) {
	
	NodeList children = list.getChildNodes();
	
	int num_nodes = children.getLength();
	String []ids = new String[num_nodes];
	for (int i=0; i<num_nodes; i++) {
	    Element e = (Element)children.item(i);
	    String id = e.getAttribute(att_name);	    
	    ids[i] = id;	    
	}
	
	return ids;
    }

    /** takes a paramList element, and gets a HashMap of name-value pairs 
     * if deep=true, extracts embedded params, otherwise just top level 
     * params*/
    public static HashMap extractParams(Element xml, boolean deep) {

	if (!xml.getNodeName().equals(PARAM_ELEM+LIST_MODIFIER)) {
	    System.err.println("GSXML:paramList element should have been passed to extractParams, instead it was "+xml.getNodeName());
	    return null;
	}

	NodeList params = null;
	if (deep) { // get all the nested ones
	    params = xml.getElementsByTagName(PARAM_ELEM);
	} else { // just get the top  level ones
	    params = xml.getChildNodes();
	}
	HashMap param_map = new HashMap();
	for (int i=0; i<params.getLength(); i++) {
	    Element param = (Element)params.item(i);
	    String name=param.getAttribute(NAME_ATT);
	    System.out.println("Extracting param "+name);
	    String value=getValue(param); //att or content
	    int pos = name.indexOf('.');
	    if (pos == -1) { // a base param
		param_map.put(name, value);
	    } else { // a namespaced param
		
		String namespace = name.substring(0, pos);
		name = name.substring(pos+1);
		System.out.println("namespace = "+namespace+" name= "+name);
		HashMap map = (HashMap)param_map.get(namespace);
		if (map == null) {
		    map = new HashMap();
		    param_map.put(namespace, map);
		}
		map.put(name, value);
	    }
	}
	return param_map;
    }

    /** gets the value att or the text content */
    public static String getValue(Element e) {
	String val = e.getAttribute(VALUE_ATT);
	if (val ==null || val.equals("")) {
	    // have to get it out of the text
	    val=getNodeText(e); 

	}
	return val;
    }

    /** extracts the text out of a node */
    public static Node getNodeTextNode(Element param) {
	param.normalize();
	Node n = param.getFirstChild();
	while (n!=null && n.getNodeType() !=Node.TEXT_NODE) {
	    n=n.getNextSibling();
	}
	return n;
    }
    
    /** extracts the text out of a node */
    public static String getNodeText(Element param) {
	Node text_node = getNodeTextNode(param);
	if (text_node == null) {
	    return "";
	}
	return text_node.getNodeValue();
    }

	
    /** add text to a document/subsection  element */
    public static boolean addDocText(Document owner, Element doc, String text) {

	Element content = owner.createElement(NODE_CONTENT_ELEM);
	Text t = owner.createTextNode(text);
	content.appendChild(t);
	doc.appendChild(content);
	return true;
    }
  
    /** add an error message */
    public static boolean addError(Document owner, Element doc, String text) {

	Element content = owner.createElement(ERROR_ELEM);
	Text t = owner.createTextNode(text);
	content.appendChild(t);

	return true;
    }

    /** add an error message */
    public static boolean addError(Document owner, Element doc, Throwable error) {
      error.printStackTrace();
      return addError(owner, doc, error.toString());
    }

    public static Element createMetadataParamList(Document owner, Vector meta_values) {

	Element meta_param_list = owner.createElement(PARAM_ELEM+LIST_MODIFIER);
	Iterator i = meta_values.iterator();
	while(i.hasNext()) {
	    String next = (String)i.next();
	    Element meta_param = owner.createElement(PARAM_ELEM);
	    meta_param_list.appendChild(meta_param);
	    meta_param.setAttribute(NAME_ATT, "metadata");
	    meta_param.setAttribute(VALUE_ATT, next);
	}
	return meta_param_list;
    }
	
    /** adds a metadata elem to a list */
    public static boolean addMetadata(Document owner, Element list, 
				      String meta_name, String meta_value) {
	if (meta_value==null || meta_value.equals("")) {
	    return false;
	}
	Element data = owner.createElement(METADATA_ELEM);
	data.setAttribute(NAME_ATT, meta_name);
	Text t = owner.createTextNode(meta_value);
	data.appendChild(t);
	list.appendChild(data);
	return true;

    }

    /** copies the metadata out of teh metadataList of 'from' into
     * the metadataList of 'to' */
    public static boolean mergeMetadataLists(Node to, Node from) {
	Node to_meta = getChildByTagName(to, METADATA_ELEM+LIST_MODIFIER);
	Node from_meta = getChildByTagName(from, METADATA_ELEM+LIST_MODIFIER);
	
	if  (from_meta == null) { // nothing to copy
	    return true;
	}
	Document to_owner = to.getOwnerDocument();
	Node new_from = to_owner.importNode(from_meta, true);

	if (to_meta == null) { // just copy the whole list
	    to.appendChild(new_from);
	    return true;
	}
	
	// copy individual elements
	Node child = new_from.getFirstChild();
	while ( child != null) {
	    to_meta.appendChild(child);
	    child = child.getNextSibling();
	}
	return true;
    }
	
    /** copies all the children from from to to */
    public static boolean mergeElements(Element to, Element from) {

	Document owner = to.getOwnerDocument();
	Node child = from.getFirstChild();
	while (child != null) {
	    to.appendChild(owner.importNode(child, true));
	    child = child.getNextSibling();
	}
	return true;
    }
    /** returns the (first) child element with the given name */
    public static Node getChildByTagName(Node n, String name) {

	Node child = n.getFirstChild();
	while (child!=null) {
	    if (child.getNodeName().equals(name)) {
		return child;
	    }
	    child = child.getNextSibling();
	}
	return null; //not found
    }

    /** takes an xpath type expression of the form name/name/... 
	and returns the first node that matches, or null if not found */
    public static Node getNodeByPath(Node n, String path) {
	
	String link = GSPath.getFirstLink(path);
	path = GSPath.removeFirstLink(path);
	while (!link.equals("")) {
	    n = getChildByTagName(n, link);
	    if (n==null) {
		return null;
	    }
	    link = GSPath.getFirstLink(path);
	    path = GSPath.removeFirstLink(path);
	}
	return n;
    }
    public static HashMap getChildrenMap(Node n) {
	
	HashMap map= new HashMap();
	Node child = n.getFirstChild();
	while (child!=null) {
	    String name = child.getNodeName();
	    map.put(name, child);
	    child = child.getNextSibling();
	}
	return map;
    }
       

    /** Duplicates an element, but gives it a new name */
    public static Element duplicateWithNewName(Document owner, Element element,
					       String element_name, boolean with_attributes)
    {
	Element duplicate = owner.createElement(element_name);

	// Copy element attributes
	if (with_attributes) {
	    NamedNodeMap attributes = element.getAttributes();
	    for (int i = 0; i < attributes.getLength(); i++) {
		Node attribute = attributes.item(i);
		duplicate.setAttribute(attribute.getNodeName(), attribute.getNodeValue());
	    }
	}

	// Copy element children
	NodeList children = element.getChildNodes();
	for (int i = 0; i < children.getLength(); i++) {
	    Node child = children.item(i);
	    duplicate.appendChild(owner.importNode(child, true));
	}

	return duplicate;
    }

    public static void copyAllChildren(Element to, Element from) {

	Document to_doc = to.getOwnerDocument();
	Node child = from.getFirstChild();
	while (child != null) {
	    to.appendChild(to_doc.importNode(child, true));
	    child = child.getNextSibling();
	}
    }
    /** returns a basic request message */
    public static  Element createBasicRequest(Document owner, 
					      String request_type, String to, 
					      String lang) {
	Element request = owner.createElement(REQUEST_ELEM);
	request.setAttribute(TYPE_ATT, request_type);
	request.setAttribute(LANG_ATT, lang);
	request.setAttribute(TO_ATT, to);
	return request;
    }
    
    public static Element createTextElement(Document owner, String elem_name,
					    String text) {
	Element e = owner.createElement(elem_name);
	Text t = owner.createTextNode(text);
	e.appendChild(t);
	return e;

    }

    public static Element createDisplayTextElement(Document owner, 
						   String text_name,
						   String text) {
	Element e = owner.createElement(DISPLAY_TEXT_ELEM);
	e.setAttribute(NAME_ATT, text_name);
	Text t = owner.createTextNode(text);
	e.appendChild(t);
	return e;

    }


    public static Element createParameter(Document owner, String name, 
					  String value) {
	Element param = owner.createElement(PARAM_ELEM);
	param.setAttribute(NAME_ATT, name);
	param.setAttribute(VALUE_ATT, value);
	return param;
    }

    public static Element createParameterList(Document owner, 
					      HashMap params) {

	Element list = owner.createElement(PARAM_ELEM+LIST_MODIFIER);
	Set items = params.entrySet();
	Iterator i = items.iterator();
	while(i.hasNext()) {
	    Map.Entry m = (Map.Entry)i.next();
	    list.appendChild(createParameter(owner, (String)m.getKey(), (String)m.getValue()));
	}
	return list;
    }
    
    public static Element createParameterDescription(Document owner, 
						     String id,
						     String display_name,
						     String type, 
						     String default_value,
						     String []option_ids,
						     String []option_names) {

	
	Element p = owner.createElement(PARAM_ELEM);
	p.setAttribute(NAME_ATT, id);
	p.setAttribute(TYPE_ATT, type);
	p.appendChild(createDisplayTextElement(owner, GSXML.DISPLAY_TEXT_NAME, display_name));
	if (default_value != null) {
	    p.setAttribute(DEFAULT_ATT, default_value);
	}
	if (option_ids!=null && option_names!=null) {
	    for (int i=0; i<option_ids.length; i++) {
		Element e = owner.createElement(PARAM_OPTION_ELEM);
		e.setAttribute(NAME_ATT, option_ids[i]);
		e.appendChild(createDisplayTextElement(owner, GSXML.DISPLAY_TEXT_NAME, option_names[i]));
		p.appendChild(e);
	    }
	}
	return p;
    }


    /** returns the element parent/node_name[@attribute_name='attribute_value']
     */
    public static Element getNamedElement(Element parent, String node_name,
					  String attribute_name, 
					  String attribute_value) {

	NodeList children = parent.getChildNodes();
	for (int i=0; i<children.getLength(); i++) {
	    Node child = children.item(i);
	    ///ystem.out.println("getnamed elem, node nmae="+child.getNodeName());
	    if (child.getNodeName().equals(node_name)) {
		if (((Element)child).getAttribute(attribute_name).equals(attribute_value))
		    return (Element)child;
	    }
	}
	// not found
	return null;
    }

   

    /** Returns the appropriate language element from a display elem,
     display is the containing element, name is the name of the element to 
     look for, lang is the preferred language, lang_default is the fall back 
     lang if neither lang is found, will return the first one it finds*/
    public static String getDisplayText(Element display, String name,
					   String lang, String lang_default) {
	
	String def = null;
	String first = null;
	NodeList elems = display.getElementsByTagName(DISPLAY_TEXT_ELEM);
	if (elems.getLength() == 0) return "";
	for (int i=0; i<elems.getLength(); i++) {
	    Element e = (Element)elems.item(i);
	    String n = e.getAttribute(NAME_ATT);
	    if (name.equals(n)) {
		String l = e.getAttribute(LANG_ATT);
		if (lang.equals(l)) {
		    return getNodeText(e);
		} else if (lang_default.equals(l)) {
			def = getNodeText(e);
		} else if (first == null) {
		    first = getNodeText(e);
		}
	    } else {
		continue;
	    }
	}

	if (def != null) {
	    return def;
	}
	if (first != null) {
	    return first;
	}
	return "";
    }
    
    // replaces < > " ' & in the original with their entities
    public static String xmlSafe(String original) {

	StringBuffer filtered = new StringBuffer(original.length());
	char c;
	for (int i=0; i<original.length(); i++) {
	    c = original.charAt(i);
	    if (c == '>') {
		filtered.append("&gt;");
	    } else if (c == '<') {
		filtered.append("&lt;");
	    } else if (c == '"') {
		filtered.append("&quot;");
	    } else if (c == '&') {
		filtered.append("&amp;");
	    } else if (c == '\'') {
		filtered.append("&apos;");
	    } else {
		filtered.append(c);
	    }
	}
	return filtered.toString();
    }

    
//      // replaces < > " ' & in the original with their entities
//      public static String unXmlSafe(String original) {

//  	StringBuffer filtered = new StringBuffer(original.length());
//  	char c;
//  	for (int i=0; i<original.length(); i++) {
//  	    c = original.charAt(i);
//  	    if (c == '&') {
//  		int pos = original.indexOf(";"), i);
//  	    String entity = original.substring(i+1, pos);
//  	    if (entity.equals("gt")) {
//  		filtered.append(">");
//  	    } else if (entity.equals("lt")) {
//  		filtered.append("<");
//  	    } else if (entity.equals("apos")) {
//  		filtered.append(
//  		while (c != ';') {
		    
//  		// process the entity
//  	    } else {
//  		filtered.append(c);
//  	    }
//  	}
//  	return filtered.toString();
//      }
}
